local StacKings = _G.StacKings

local L = LibStub("AceLocale-3.0"):GetLocale("StacKings")
local LSM = LibStub("LibSharedMedia-3.0")

local debug = StacKings.debug

local frames = {}
StacKings.frames = frames

local function formatTime(time)
	if time > 60 then
		local minutes = math.floor(time/60)
		local seconds = time - minutes * 60
		if seconds < 10 then
			return ("%d:0%d"):format(minutes, seconds)
		else
			return ("%d:%d"):format(minutes, seconds)
		end
	else
		return ("%.1f"):format(time)
	end
end

local function getAura(unit, spell, auratype)
	if auratype == "HELPFUL" then
		return UnitAura(unit, spell)
	elseif auratype == "HARMFUL" then
		return UnitDebuff(unit, spell)
	end
end

local function updateTime(self, spell)
	local value = self.value
	local unit = self.unit
	local spell = self.spell
	local _, _, _, _, _, _, exp = getAura(unit, spell, self.auratype)
	if not exp then
		value:SetText("")
		return
	end
	local rem = exp - GetTime()
	if self.showthreshold and rem > self.showthreshold then
		value:SetText("")
	else
		if self.flash and rem < 10 then
			value:SetTextColor(1,0,0) -- make it red when imminent
		end
		value:SetText(formatTime(rem)) -- show tenth
	end
end

local function isOverrideColor(rangestring, count)
	local lower, upper = ("-"):split(rangestring)
	lower = tonumber(lower)
	upper = tonumber(upper)
	if not lower then return end
	if upper then
		return count <= upper and count >= lower
	else
		return count == lower
	end
end

local function findOverrideColor(override, count)
	for _,v in ipairs(override) do
		if isOverrideColor(v.range, count) then
			return v.color
		end
	end
end

local function updateBarColor(bar, count)
	local overridecolor = findOverrideColor(bar.coloroverride, count)
	if overridecolor then
		bar:SetStatusBarColor(unpack(overridecolor))
	else
		bar:SetStatusBarColor(unpack(bar.normalcolor))
	end
end

local function auraUpdate(self)
	local bar = self.bar
	local spell = self.spell
	local unit = self.unit
	local _, _, _, count = getAura(unit, spell, self.auratype)
	if not count then
		count = 0
	end
	updateBarColor(bar, count)
	bar:SetValue(count)
end

local function makeFontString(parent)
	local fs = parent:CreateFontString(nil, "OVERLAY")
	fs:SetShadowColor(0,0,0)
	fs:SetShadowOffset(0.8, -0.8)
	return fs
end

local function makeBar(frame)
	local bar = CreateFrame('StatusBar', nil, frame)
	bar:SetAllPoints(frame)
	return bar
end
local function makeBackground(bar)
	local bg = bar:CreateTexture(nil, "BORDER")
	bg:SetAllPoints(bar)
	return bg
end

local function insertValues(table, ...)
	for i = 1,select('#', ...) do
		table[i] = select(i, ...)
	end
end

local function updatePoint(self)
	local point = self.point
	wipe(point)
	local pointname, relativeFrame, relativePoint, x, y = self:GetPoint(1)
	point[1] = pointname
	point[2] = relativeFrame
	point[3] = relativePoint
	point[4] = x
	point[5] = y
end

local function onDragStop(frame)
	frame:StopMovingOrSizing()
	frame:UpdatePoint()
end

local function setupEvents(frame)
	frame.UNIT_AURA = auraUpdate
	frame:SetScript("OnEvent", auraUpdate)
	frame:SetScript("OnUpdate", updateTime)
	frame:SetScript("OnMouseDown", frame.StartMoving)
	frame:SetScript("OnMouseUp", onDragStop)
end
local function makeValue(bar)
	local fs = makeFontString(bar)
	fs:SetPoint("CENTER", bar, "CENTER")
	return fs
end

local function setupBar(bar, settings)
	bar:SetStatusBarTexture(LSM:Fetch('statusbar', settings.bartexture))
	bar:SetStatusBarColor(unpack(settings.color))
	bar:SetMinMaxValues(0, settings.max)
	bar:SetValue(0)
	bar.normalcolor = settings.color
	bar.coloroverride = settings.coloroverride
end

local function setupBackground(bg, settings)
	bg:SetTexture(LSM:Fetch('statusbar', settings.bartexture))
	bg:SetVertexColor(unpack(settings.bgcolor))
	bg:SetAlpha(settings.bgalpha)
end

local function setupValue(value, settings)
	local font = LSM:Fetch('font', settings.fontface)
	value:SetFont(font, settings.fontsize, settings.fontoutline)
	value:SetTextColor(unpack(settings.fontcolor))
	return value
end

local function lockFrame(frame, lock)
	frame:SetMovable(lock)
	frame:EnableMouse(lock)
end

local unitChangedEvents = {
	target = "PLAYER_TARGET_CHANGED",
	focus = "PLAYER_FOCUS_CHANGED",
	pet = "PLAYER_PET_CHANGED",
}
local function setupUnitEvents(frame, unit)
	for k,v in pairs(unitChangedEvents) do
		if unit == k then
			frame:RegisterEvent(v)
		else
			frame:UnregisterEvent(v)
		end
	end
end

local function disableFrame(frame)
	if not frame then
		debug("cannot disable nil-frame")
		return
	end
	UnregisterStateDriver(frame, "visibility")
	frame:UnregisterAllEvents()
	frame:Hide()
	frame.disabled = true
end

local function enableFrame(frame)
	setupEvents(frame)
	frame.disabled = false
end

local function autoInitWithTable(self, key)
	local val = rawget(self, key)
	if not val then
		val = {}
		self[key] = val
	end
	return val
end

local delayedCalls = setmetatable({}, { __index = autoInitWithTable, })
local function callAllDelayed()
	StacKings:UnregisterEvent("PLAYER_REGEN_ENABLED")
	for func,v in pairs(delayedCalls) do
		for i, args in ipairs(v) do
			func(unpack(args))
			wipe(args)
			v[i] = nil
		end
		delayedCalls[func] = nil
	end
end

local function executeOutOfCombat(func, ...)
	if InCombatLockdown() then
		table.insert(delayedCalls[func], { ... })
		StacKings:RegisterEvent("PLAYER_REGEN_ENABLED", callAllDelayed)
	else
		func(...)
	end
end

local function setupSpecCheck(frame, value)
	if value and value ~= 0 then
		RegisterStateDriver(frame, "visibility", ("[spec:%d]show;hide"):format(value))
	else
		UnregisterStateDriver(frame, "visibility")
		frame:Show()
	end
end

local function handlePosAndSize(frame, settings)
	frame:SetHeight(settings.height)
	frame:SetWidth(settings.width)
	frame:ClearAllPoints()
	frame:SetPoint(unpack(settings.point))
	frame.point = settings.point
	lockFrame(frame, not settings.locked)
end

local function setupIcon(frame, settings)
	if settings.iconenable then
		local texture
		if not settings.icontexture then
			texture = select(3, GetSpellInfo(settings.spell))
		elseif tonumber(settings.icontexture) then
			texture = select(3, GetSpellInfo(settings.icontexture))
		else
			texture = settings.icontexture
		end

		frame.icon:SetTexture(texture)
		frame.icon:SetWidth(settings.height)
		frame.icon:SetHeight(settings.height)
		frame.icon:SetPoint("RIGHT", frame.bar, "LEFT", 0, 0)
		frame.icon:Show()
	else
		frame.icon:Hide()
	end
end

local function setupFrameVariables(frame, settings)
	frame.spell = settings.spell
	frame.unit = settings.unit
	frame.auratype = settings.auratype
	frame.showthreshold = settings.showthreshold
	frame.flash = settings.flash
end

local function setupTime(frame, settings)
	if settings.showTime then
		frame.value:Show()
		frame.showthreshold = settings.showthreshold
		frame.flash = settings.flash
	else
		frame.value:Hide()
	end
end

local function applySettings(frame, settings)
	if settings.enabled and frame.disabled then
		executeOutOfCombat(enableFrame, frame)
	elseif not settings.enabled then
		if not frame.disabled then
			executeOutOfCombat(disableFrame, frame)
		end
		return
	end

	executeOutOfCombat(handlePosAndSize, frame, settings)

	setupFrameVariables(frame, settings)
	setupBar(frame.bar, settings)
	setupBackground(frame.bg, settings)
	setupValue(frame.value, settings)
	setupTime(frame, settings)

	setupIcon(frame, settings)
	setupUnitEvents(frame, settings.unit)
	executeOutOfCombat(setupSpecCheck, frame, settings.spec)
	frame:UNIT_AURA(settings.unit)
end

local buckets = setmetatable({}, { __index = autoInitWithTable, })
local function moveToBucket(name, oldunit, unit)
	if oldunit ~= unit then
		if oldunit then
			buckets[oldunit][name] = nil
		end
		buckets[unit][name] = frames[name]
	end
end
function StacKings:UNIT_AURA(_, unit)
	for k, v in pairs(buckets[unit]) do
		v:UNIT_AURA()
	end
end

local function makeIcon(frame)
	local icon = frame:CreateTexture(nil, "BORDER")
	return icon
end

local function makeNewFrame()
	local frame = CreateFrame('Frame', nil, UIParent, "SecureHandlerStateTemplate")
	frame.bar = makeBar(frame)
	frame.bg = makeBackground(frame.bar)
	frame.value = makeValue(frame.bar)
	frame.icon = makeIcon(frame.bar)
	frame.UpdatePoint = updatePoint
	return frame
end

local framepool = {}
function StacKings:CreateBarFrame(key)
	local frame = next(framepool)
	if frame then
		framepool[frame] = nil
	else
		frame = makeNewFrame()
	end
	setupEvents(frame)
	frames[key] = frame
	return frame
end

function StacKings:RemoveBarFrame(name)
	local frame = frames[name]
	frames[name] = nil
	disableFrame(frame)
	framepool[frame] = true
end

function StacKings:ApplyFrameSettings(name, settings)
	local frame = frames[name]
	moveToBucket(name, frame.unit, settings.unit)
	applySettings(frame, settings)
end

function StacKings:InitializeBars()
	self:RegisterEvent("UNIT_AURA")
	for k,v in pairs(self.db.profile.bars) do
		local frame = self:CreateBarFrame(k)
		self:ApplyFrameSettings(k, v)
	end
end

function StacKings:UpdateAllBars()
	for k,v in pairs(frames) do
		v:UNIT_AURA()
	end
end